////////////////////////////////////////////////////////////////////////////
// #02012633
//
// Copyright (C) 2002 Albert Vik
//
// E-mail: AlbertVik@operamail.com
//
// This source code is part of FLAC project. This source code is distributed 
// under the same policy as the FLAC project constitute except one addition:
// any modifications of this file and the source code in this file for 
// purposes other than for private use should be sanctioned by the author.
//
// The latest release of this source code is available on the  
// FLAC project home page: http://flac.sourceforge.net or
// directly from the author: AlbertVik@operamail.com
// 
////////////////////////////////////////////////////////////////////////////

#include "FLAC_Apollo.h"

///////////////////////////////////////////////////////////////////////////
//
// Class CFLACDecoder - FLAC Decoding Functionality
//
///////////////////////////////////////////////////////////////////////////

CFLACDecoder::CFLACDecoder( char *filename, int audioDataOffset )
{
    if (strstr(filename,"://") != NULL) return;    // No URLs

    if( !GetInfo(filename) ) return;

    m_AbortFlag = FALSE;

    m_decoder = FLAC__file_decoder_new();

    if( !m_decoder ) return; // ERROR creating the decoder instance

    // sizeof(FLAC__int32) used instead of (FLAC__MAX_BITS_PER_SAMPLE+7)/8
    // because array FLAC__int32 used even for 24 bps
    
    m_BufferSize = MaxBlocksize() * Channels() * sizeof(FLAC__int32);
    m_ProcessedBuffer = new FLAC__int8 [m_BufferSize];

    FLAC__file_decoder_set_md5_checking(m_decoder, TRUE);
    FLAC__file_decoder_set_filename(m_decoder, filename);
	
    FLAC__file_decoder_set_write_callback    (m_decoder, (FLAC__StreamDecoderWriteStatus (*)(const FLAC__FileDecoder *, const FLAC__Frame *, const FLAC__int32 *[], void *))flac_callback_write);
    FLAC__file_decoder_set_metadata_callback (m_decoder, (void (*)(const FLAC__FileDecoder *,                           const FLAC__StreamMetaData *,   void *))            flac_callback_metadata);
    FLAC__file_decoder_set_error_callback    (m_decoder, (void (*)(const FLAC__FileDecoder *,                           FLAC__StreamDecoderErrorStatus, void *))            flac_callback_error);
	
    FLAC__file_decoder_set_client_data(m_decoder, this);

    FLAC__FileDecoderState state = FLAC__file_decoder_init(m_decoder);

    if( state != FLAC__FILE_DECODER_OK )  // ERROR initializing decoder
    {
        m_AbortFlag = TRUE;
        return; 
    }

    if( !FLAC__file_decoder_process_metadata( m_decoder ) ) 
    {                                     // ERROR processing metadata
        m_AbortFlag = TRUE;
        return; 
    }

    return;

} // CFLACDecoder()



CFLACDecoder::~CFLACDecoder()
{
    if( m_decoder ) 
    {
       FLAC__file_decoder_finish( m_decoder );
       FLAC__file_decoder_delete( m_decoder );
    }

    if( m_ProcessedBuffer != NULL )  delete [] m_ProcessedBuffer;

    return;

}  //  CFLACDecoder::~CFLACDecoder()



int CFLACDecoder::GetNextAudioChunk(void **buffer)
{
    // To avoid writing waste chunks while decoder searches for EOF
    m_SamplesInBuffer = 0;  
	
    if( !FLAC__file_decoder_process_one_frame(m_decoder) ) 
    {
        return -1;   // Returning -1 means error in reading data
    }

    FLAC__FileDecoderState state = FLAC__file_decoder_get_state( m_decoder );
    
    if( state == FLAC__FILE_DECODER_END_OF_FILE ) 
    {
        return 0;    // Returning 0 means end of data
    }
     
    DWORD numBytes = m_SamplesInBuffer * SampleWidth();
	
    // The size of decoded data may vary from call to call, but not over 50K

    if( numBytes > 50000 ) // Probably Apollo mean 51200, but let's not risk
    {
        return -1;   // Returning -1 means error in reading data
    }

    // Set the buffer-parameter to point to uncompressed data.
    // The Apollo player will not manipulate this data.

    *buffer = m_ProcessedBuffer;

    return numBytes;   // Return the data size in bytes

}  // CFLACDecoder::GetNextAudioChunk()



int CFLACDecoder::SetPosition( int newPosition )
{
    unsigned position;   // Milliseconds

    if( newPosition < 0)  position = 0;
    else                  position = newPosition;

    if( position > PlayingTime() ) position = PlayingTime();

    unsigned target_sample = (unsigned)(
                             (double)position * 
                             (double)TotalSamples() / 
                             (double)PlayingTime()  );
                    
    FLAC__file_decoder_seek_absolute( m_decoder, target_sample );

    return position;

} // CFLACDecoder::SetPosition()



/////////////////////////////////////////////////////////////////////////////
// libFLAC metadata callback - This method will be called by libFLAC 
// through the appropriate global stub
//
void CFLACDecoder::callback_metadata( const FLAC__StreamMetaData *metadata )
{
    if(metadata->type == FLAC__METADATA_TYPE_STREAMINFO)
    {
        if( BitsPerSample() != 8 && BitsPerSample() != 16 && BitsPerSample() != 24)
        {
            m_AbortFlag = TRUE;
            return;  // ERROR: bits per sample is not 8/16/24
        }
    }

    return;

}  // CFLACDecoder::callback_metadata()



/////////////////////////////////////////////////////////////////////////////
// libFLAC write callback - This method will be called through
// the appropriate global stub when the decoder has decoded a 
// single frame of data. The decoder will pass the frame 
// metadata as well as an array of pointers (one for each channel) 
// pointing to the decoded audio.
//
FLAC__StreamDecoderWriteStatus CFLACDecoder::callback_write( const FLAC__Frame *frame, const FLAC__int32 *buffer[] )
{
    if( m_AbortFlag )  return FLAC__STREAM_DECODER_WRITE_ABORT;

    FLAC__int8   *s8buffer  = (FLAC__int8   *)m_ProcessedBuffer;
    FLAC__int16  *s16buffer = (FLAC__int16  *)m_ProcessedBuffer;
    FLAC__int32  *s32buffer = (FLAC__int32  *)m_ProcessedBuffer;

    unsigned wide_sample, sample, channel;

    unsigned wide_samples = frame->header.blocksize;

    if(BitsPerSample() == 8) 
    {
        for(sample = wide_sample = 0; wide_sample < wide_samples; wide_sample++)
            for(channel = 0; channel < Channels(); channel++, sample++)
                s8buffer[sample] = (FLAC__int8)(buffer[channel][wide_sample]);
    }
    else if(BitsPerSample() == 16) 
    {
        for(sample = wide_sample = 0; wide_sample < wide_samples; wide_sample++)
            for(channel = 0; channel < Channels(); channel++, sample++)
                s16buffer[sample] = (FLAC__int16)(buffer[channel][wide_sample]);
    }
    else if(BitsPerSample() == 24)
    {
        for(sample = wide_sample = 0; wide_sample < wide_samples; wide_sample++)
            for(channel = 0; channel < Channels(); channel++, sample++)
                s32buffer[sample] = buffer[channel][wide_sample];

        unsigned byte, lbyte;
        const unsigned bytes = sample * 4;
        for(lbyte = byte = 0; byte < bytes; ) 
        { 
            s8buffer[lbyte++] = s8buffer[byte++];
            s8buffer[lbyte++] = s8buffer[byte++];
            s8buffer[lbyte++] = s8buffer[byte++];
            byte++;
        }
    }
    else FLAC__ASSERT(0);

    m_SamplesInBuffer = sample;

    return FLAC__STREAM_DECODER_WRITE_CONTINUE;

}  // CFLACDecoder::callback_write()



/////////////////////////////////////////////////////////////////////////////
// libFLAC error callback - This method will be called by libFLAC through
// the appropriate global stub
//
void CFLACDecoder::callback_error( FLAC__StreamDecoderErrorStatus status )
{
    const char *error = FLAC__StreamDecoderErrorStatusString[status];

    // NOTE: Subject for plugin Configure:
    // set abort_flag to TRUE for stop playing on error

    m_AbortFlag = FALSE;  // Hardcoded for ignore audio data corruption.

}  // CFLACDecoder::callback_error()



////////////////////////////////////////////////////////////////////////////
//
// Global libFLAC Callback Stubs
//
////////////////////////////////////////////////////////////////////////////

void flac_callback_metadata(const FLAC__FileDecoder *decoder, const FLAC__StreamMetaData *metadata, void *client_data)
{
    if( decoder == ((CFLACDecoder*)client_data)->GetDecoder() )
        ((CFLACDecoder*)client_data)->callback_metadata( metadata );

}  // ::flac_callback_metadata()



FLAC__StreamDecoderWriteStatus flac_callback_write(const FLAC__FileDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 *buffer[], void *client_data)
{
    if( decoder == ((CFLACDecoder*)client_data)->GetDecoder() ) 
        return ((CFLACDecoder*)client_data)->callback_write( frame, buffer );
    else
		return FLAC__STREAM_DECODER_WRITE_ABORT;

}  // ::flac_callback_write()



void flac_callback_error(const FLAC__FileDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data)
{
    if( decoder == ((CFLACDecoder*)client_data)->GetDecoder() ) 
        ((CFLACDecoder*)client_data)->callback_error( status );

}  // ::flac_callback_error()

